package ru.dbotthepony.kstarbound.client.gl.program

import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.shader.GLShader
import ru.dbotthepony.kstarbound.client.gl.shader.GLTransformableColorableProgram
import ru.dbotthepony.kstarbound.client.gl.shader.GLTransformableProgram
import ru.dbotthepony.kstarbound.client.gl.vertex.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.VertexType
import kotlin.properties.ReadOnlyProperty
import kotlin.reflect.KProperty

private class SimpleProgram<T : GLShaderProgram>(private val name: String, private val factory: (GLStateTracker, Array<out GLShader>) -> T) : ReadOnlyProperty<GLPrograms, T> {
	private var value: T? = null

	override fun getValue(thisRef: GLPrograms, property: KProperty<*>): T {
		thisRef.state.ensureSameThread()

		val value = value

		if (value != null) {
			return value
		}

		val vertex = GLShader.internalVertex("shaders/$name.vsh")
		val fragment = GLShader.internalFragment("shaders/$name.fsh")

		val newValue = factory.invoke(thisRef.state, arrayOf(vertex, fragment))

		this.value = newValue

		vertex.unlink()
		fragment.unlink()

		if (!newValue.linked) {
			newValue.link()
		}

		return newValue
	}
}

class GLLiquidProgram(state: GLStateTracker) : GLInternalProgram(state, "liquid") {
	init {
		link()
	}

	val baselineColor = this["baselineColor"]!!
	val transform = this["transform"]!!

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, VertexType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
	}
}

class GLPrograms(val state: GLStateTracker) {
	val tile by SimpleProgram("tile", ::GLTransformableColorableProgram)
	val font by SimpleProgram("font", ::GLTransformableColorableProgram)
	val flat by SimpleProgram("flat", ::GLTransformableColorableProgram)
	val liquid by lazy { GLLiquidProgram(state) }
}
